#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Codegeneration utilities.
"""

# Copyright (C) 2008-2009 Martin Sandve Alnes and Simula Resarch Laboratory
#
# This file is part of SyFi.
#
# SyFi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# SyFi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SyFi. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Kent-Andre Mardal, 2010.
#
# First added:  2008-08-13
# Last changed: 2009-05-15

from itertools import chain
import re
import ufl
import ufc_utils
import SyFi

from ufl.classes import Measure
from ufl.algorithms import FormData, preprocess

from dolfin_utils.wrappers import generate_dolfin_code, UFCFormNames
import sfc
from sfc.common.utilities import unique
from sfc.common.output import sfc_assert, sfc_error, sfc_info, sfc_debug
#from sfc.common import ParameterDict
from sfc.common.options import default_parameters

from sfc.representation import GeometryRepresentation
from sfc.representation import ElementRepresentation
from sfc.representation import FormRepresentation

from sfc.codegeneration.formcg import FormCG
from sfc.codegeneration.dofmapcg import DofMapCG
from sfc.codegeneration.finiteelementcg import FiniteElementCG

# FIXME: Move these somewhere else
def integral_handlers(domaintype, options):
    # Switch family of classes based on options
    if options.code.integral.use_uflacs:
        import sfc.representation.integralrep_uflacs as r
        import sfc.codegeneration.integralcg_uflacs as c
    else:
        import sfc.representation.integralrep_regular as r
        import sfc.codegeneration.integralcg_regular as c

    # Select integral classes based on domain type
    if domaintype == Measure.CELL:
        return (r.CellIntegralRepresentation,
                c.CellIntegralCG)
    if domaintype == Measure.EXTERIOR_FACET:
        return (r.ExteriorFacetIntegralRepresentation,
                c.ExteriorFacetIntegralCG)
    if domaintype == Measure.INTERIOR_FACET:
        return (r.InteriorFacetIntegralRepresentation,
                c.InteriorFacetIntegralCG)


# version %(version)s
_header_template = r"""/*
 * %(name)s.h
 *
 * This file was automatically generated by SFC.
 *
 * http://www.fenics.org/syfi/
 *
 */

#ifndef GENERATED_SFC_CODE_%(name)s
#define GENERATED_SFC_CODE_%(name)s

#include "ufc.h"

%(hincludes)s

//namespace sfc
//{
%(hbody)s
//}

#endif
"""

# version %(version)s
_implementation_template = r"""/*
 * %(name)s.cpp
 *
 * This file was automatically generated by SFC.
 *
 * http://www.fenics.org/syfi/
 *
 */

#include "ufc.h"

#include "%(name)s.h"

%(cincludes)s

//namespace // Anonymous namespace
//{
%(supportcode)s
//}

//namespace sfc
//{
%(cbody)s
//}
"""

def common_system_includes():
    return ("iostream", "cmath", "stdexcept", "cstring")

def apply_code_dict(format_string, code_dict):
    "Template formatting with improved checking for template argument mismatch."
    expected_keys = set(re.findall('%\((\w+)\)', format_string))
    actual_keys = set(code_dict.keys())

    missing_keys = expected_keys.difference(actual_keys)
    if missing_keys:
        print "When formatting code using template, missing keys:"
        print "\n".join(sorted(missing_keys))
        print

    #superfluous_keys = actual_keys.difference(expected_keys)
    #if superfluous_keys:
    #    print "When formatting code using template, got superfluous keys:"
    #    print "\n".join(sorted(superfluous_keys))
    #    print

    return format_string % code_dict

def generate_class_files(cg):
    sfc_debug("Entering generate_class_files")

    code_dict = cg.generate_code_dict()
    classname = code_dict["classname"]

    hbody = apply_code_dict(cg.header_template(), code_dict)
    supportcode = cg.generate_support_code()
    cbody = apply_code_dict(cg.impl_template(), code_dict)

    # TODO: Probably don't need all these always, could speed up compilation a bit
    system_includes = common_system_includes()

    hincludes = "\n".join('#include "%s"' % inc for inc in cg.hincludes())

    cincludes =  "\n".join('#include <%s>' % f for f in system_includes)
    cincludes += "\n"
    cincludes += "\n".join('#include "%s"' % inc
                           for inc in chain(cg.cincludes(), cg.local_includes()))

    hdict = { "name": classname,
              "hbody": hbody,
              "hincludes": hincludes }
    cdict = { "name": classname,
              "supportcode": supportcode,
              "cbody": cbody,
              "cincludes": cincludes }

    hcode = _header_template % hdict
    ccode = _implementation_template % cdict

    # External includes, i.e. ones not generated by sfc
    includes = cg.hincludes() + cg.cincludes()

    sfc_debug("Leaving generate_class_files")
    return classname, (hcode, ccode), includes

def write_file(filename, contents):
    "Write contents to a file."
    #print "WRITING FILE", filename
    # TODO: Setup some kind of hook here for testing without writing files?
    f = open(filename, "w")
    f.write(contents)
    f.close()

def write_code(classname, code):
    "Write code for a class to a header file or header + impl file."
    sfc_debug("Entering write_code")
    if isinstance(code, tuple):
        # Code is split in a header and implementation file
        hcode, ccode = code
        hname = classname + ".h"
        cname = classname + ".cpp"
        write_file(hname, hcode)
        write_file(cname, ccode)
        ret = (hname, cname)
    else:
        # All code is combined in a header file
        name = classname + ".h"
        write_file(name, code)
        ret = (name,)
    sfc_debug("Leaving write_code")
    return ret

def compiler_input(input, objects=None, common_cell=None):
    """Map different kinds of input to a list of
    UFL elements and a list of FormData instances.

    The following input formats are allowed:
    - ufl.Form
    - ufl.algorithms.FormData
    - ufl.FiniteElementBase
    - list of the above

    Returns:
        elements, formdatas
    """
    sfc_debug("Entering compiler_input")
    sfc_assert(input, "Got no input!")
    fd = []
    fe = []
    if not isinstance(input, list):
        input = [input]
    for d in input:
        if isinstance(d, ufl.form.Form):
            d = d.compute_form_data(object_names=objects, common_cell=common_cell)
        if isinstance(d, ufl.algorithms.formdata.FormData):
            fd.append(d)
            fe.extend(d.sub_elements)
        elif isinstance(d, ufl.FiniteElementBase):
            if d.cell().is_undefined():
                sfc_assert(common_cell and not common_cell.is_undefined(),
                           "Element no defined cell, cannot compile this.")
                d = d.reconstruct(cell=common_cell)
                sfc_assert(not d.cell().is_undefined(), "Still undefined?")
            fe.append(d)
        else:
            sfc_error("Not a FormData or FiniteElementBase object: %s" % str(type(d)))
    fe = sorted(set(fe))
    sfc_debug("Leaving compiler_input")
    for x in fe:
        sfc_assert(not x.cell().is_undefined(),
                   "Can never have undefined cells at this point!")
    for x in fd:
        sfc_assert(not x.cell.is_undefined(),
                   "Can never have undefined cells at this point!")
    return fe, fd

dolfin_header_template = """/*
 * DOLFIN wrapper code generated by the SyFi Form Compiler.
 */

%s
"""
def generate_dolfin_header_file(prefix, formnames, hfilenames):
    if formnames:
        sfc_info("Generating DOLFIN wrapper code, formnames are %s." %\
                 "\n".join(map(str,formnames)))

        header = dolfin_header_template % "\n".join('#include "%s"' % h
                                                    for h in hfilenames)
        dolfin_code = generate_dolfin_code(prefix, header, formnames)

        dolfin_filename = "%s.h" % prefix
        write_file(dolfin_filename, dolfin_code)
        return dolfin_filename
    else:
        # TODO: Generate FunctionSpaces for elements?
        sfc_info("NOT generating dolfin wrappers, no forms to wrap!")
        return None

def update_global_syfi_nsd(nsd):
    if nsd == update_global_syfi_nsd.last_nsd:
        return # TODO: Check actual dimension?
    else:
        update_global_syfi_nsd.last_nsd = nsd
        SyFi.initSyFi(nsd)
update_global_syfi_nsd.last_nsd = None

def update_element_representations(e, geomrep, element_reps, generated_elements, options):
    # Already visited this element? Skip it if possible.
    if e in element_reps:
        return []

    # Construct ElementRepresentation objects for all elements
    quad_rule = None # FIXME: What to do with this one?
    assert not "Quadrature" in e.family() # No quadrature rule defined!

    # Create representation and cache for reuse
    erep = ElementRepresentation(e, geomrep, quad_rule=quad_rule,
                                 options=options, cache=element_reps)
    element_reps[e] = erep

    # Build flat list of all subelements
    todo = [erep]
    ereps = []
    while todo:
        erep = todo.pop()
        # Skip already generated subelements!
        if not erep.ufl_element in generated_elements:
            generated_elements.add(erep.ufl_element)
            ereps.append(erep)
        # Queue subelements for inspection
        for subrep in erep.sub_elements:
            todo.append(subrep)

    return ereps

def generate_element_ufc_code(element_reps_list, options):
    # Generate code for all element representations
    generated_classes = []
    for rep in element_reps_list:
        # Generate code for finite_element
        cg = FiniteElementCG(rep)
        generated_classes.append(generate_class_files(cg))

        # Generate code for dof_map
        cg = DofMapCG(rep)
        generated_classes.append(generate_class_files(cg))
    return generated_classes

def generate_form_ufc_code(formrep, options):

    pf = formdata = formrep.formdata.preprocessed_form
    ig = pf.integral_groups()

    generated_classes = []

    # Generate code for all types of integral classes
    for domaintype in (Measure.CELL, Measure.EXTERIOR_FACET, Measure.INTERIOR_FACET):
        # Generate code for this particular integral type
        IntegralRepresentation, IntegralCG = integral_handlers(domaintype, options)

        for subdomain in pf.domains(domaintype):
            integrals = ig[subdomain]
            if integrals:
                # Create symbolic representation for this integral
                itgrep = IntegralRepresentation(integrals, formrep)

                # Create code generation object for this integral class
                cg = IntegralCG(itgrep)

                # Generate code for this class
                generated_classes.append(generate_class_files(cg))

    # Generate code for form class, which basically collects everything
    cg = FormCG(formrep)
    generated_classes.append(generate_class_files(cg))

    return generated_classes

def create_geometry_rep(domain, geometry_reps):
    nsd = ufl.geometry.domain2dim[domain]

    # Initialize global variables in SyFi with the
    # right space dimension (not very nice! need a
    # significant refactoring of SyFi to fix this...)
    update_global_syfi_nsd(nsd)

    if not domain in geometry_reps:
        geomrep = GeometryRepresentation(domain)
        geometry_reps[domain] = geomrep

    return geometry_reps[domain]

def generate_code(input, objects, options=None, common_cell=None):
    """Generate code from input and options.

    @param input:
        TODO
    @param options:
        TODO
    """
    sfc_debug("Entering generate_code")
    sfc_assert(input, "Got no input!")

    if options is None:
        options = default_parameters()

    DEBUG = 0
    if DEBUG: print "Calling compiler_input"
    ufl_elements, formdatas = compiler_input(input, objects, common_cell=common_cell)

    generated_classes = []
    geometry_reps = {}
    element_reps = {}
    generated_elements = set()

    # Generate UFC code for elements
    if DEBUG: print "Making element representations"
    for e in ufl_elements:
        # Create a shared geometry representation
        # (sets the global space dimension in SyFi as a side effect)
        geomrep = create_geometry_rep(e.cell().domain(), geometry_reps)

        # Create symbolic representations for all new elements
        element_reps_list = update_element_representations(e,
                                                           geomrep,
                                                           element_reps,
                                                           generated_elements,
                                                           options)

        # Generate code for all new elements (finite_element and dof_map)
        gc = generate_element_ufc_code(element_reps_list, options)
        generated_classes.extend(gc)

    # Generate UFC code for forms
    formnames = []
    for formdata in formdatas:
        if DEBUG: print "Starting processing another form."

        # Create a shared geometry representation
        # (sets the global space dimension in SyFi as a side effect)
        geomrep = create_geometry_rep(formdata.cell.domain(), geometry_reps)

        # Build symbolic representation
        if DEBUG: print "Making form representation"
        formrep = FormRepresentation(formdata, element_reps, geomrep, options)

        # Generate code for form class
        if DEBUG: print "Generating form code"
        gc = generate_form_ufc_code(formrep, options)
        generated_classes.extend(gc)

        # Collect classnames for use with dolfin wrappers
        namespace = "" # "sfc::"
        fn = UFCFormNames(formdata.name,
                          formdata.coefficient_names,
                          namespace + formrep.classname,
                          [namespace + name for name in formrep.fe_names],
                          [namespace + name for name in formrep.dm_names])
        formnames.append(fn)

        if DEBUG: print "Done with this form."

    # Write code to files
    filenames = []
    needed_files = []
    for classname, code, includes in generated_classes:
        fn = write_code(classname, code)
        filenames.extend(fn)
        needed_files.extend(includes)

    # Fetch other needed files:
    if needed_files:
        raise NotImplementedError("FIXME: Implement fetching non-ufc-class "+\
                                  "files like DofPtv and quadrature rule files.")

    filenames = list(unique(chain(filenames, needed_files)))
    hfilenames = [f for f in filenames if f.endswith(".h")]
    cfilenames = [f for f in filenames if f.endswith(".cpp")]

    # Generate DOLFIN wrapper code
    if options.code.dolfin_wrappers:
        hfn = generate_dolfin_header_file(options.code.prefix, formnames, hfilenames)
        if hfn: hfilenames.append(hfn)

    sfc_debug("Leaving generate_code")
    return hfilenames, cfilenames
