
from ufl.common import component_to_index, product
from ufl.permutation import compute_indices

"""
FIXME:

Get information:

- List of required elements with number of derivatives needed

- Element representation for all required elements

- Element representation for all required SUBelements

- Mappings from elements to unique subelements


Without derivatives:

- Generate array members for precompute struct

- Fill array members of precompute struct

- Use precomputed basis functions in argument loops

- Use linear combinations of dofs and basis functions for coefficients


With derivatives:

- Generate array members for local gradients for precompute struct

- Fill array members of precompute struct with local gradients

- Use precomputed basis function gradients with applied geometry mapping in argument loops

- Use linear combinations of dofs and mapped basis function gradients for coefficients

"""

import ufl
from uflacs.codeutils.target_formatter import analyse_terminalish
from uflacs.codeutils.format_code_structure import (format_code_structure,
                                                    ArrayDecl, ArrayAccess)

def Assignment(lhs, rhs): # TODO: Implement in uflacs
    return (lhs, " = ", rhs, ";")

class FormArgumentCG(object):
    def __init__(self, farep, on_facet):
        self.farep = farep
        self.geomrep = farep.geomrep
        self.on_facet = on_facet

    def precompute(self, terminalish, required):
        self.pcelm = []

        fr = self.farep.formrep

        # Figure out how many derivatives we need for which element
        todo = set()
        for o in terminalish:
            t, ngrads, r = analyse_terminalish(o)

            if isinstance(t, ufl.Argument):
                count = t.count()
            elif isinstance(t, ufl.Coefficient):
                count = t.count() + fr.rank
            else:
                count = None

            if count is not None:
                elm = fr.formdata.elements[count]
                todo.add((elm, ngrads, r))

        #prefix_template = "pc[%(iq)s]."
        #phi_template = "%(diff)sphi%(elmnum)d[%(idof)s][%(icomp)s]"

        # Loop over element representations and emit precomputation code
        for elm, ngrads, r in todo:
            assert not r # FIXME
            assert not ngrads # FIXME
            er = fr.element_reps[elm]

            # Handle both mixed and simple elements same way,
            # keeps it generic for now, some optimizations lost

            #name = "d%d_phi%d" % (0, 0,) # FIXME: Need unique element count here

            name = "phi%d" % (0,) # FIXME: Need unique element count here
            shape = (er.local_dimension, er.value_size)
            decl = ArrayDecl("double", name, shape)
            data = decl, er, ngrads, r
            self.pcelm.append(data)

    def _scratch(self):
        if 0:
            if isinstance(elm, ufl.MixedElement):
                assert False # FIXME: Support arbitrary mixed elements
                # Check for unique subelements of er and add to todo:
                for ser in er.sub_elements: # TODO
                    #name = "d%d_phi%d" % (0, 0,) # FIXME: Need unique element count here
                    name = "phi%d" % (0,) # FIXME: Need unique element count here
                    shape = (er.local_dimension, er.value_size) # FIXME: symmetries? other?
                    decl = ArrayDecl("double", name, shape)
                    data = decl, ser, ngrads, r
                    self.pcelm.append(data)
            else:
                #name = "d%d_phi%d" % (0, 0,) # FIXME: Need unique element count here
                name = "phi%d" % (0,) # FIXME: Need unique element count here
                shape = (er.local_dimension, er.value_size)
                decl = ArrayDecl("double", name, shape)
                data = decl, er, ngrads, r
                self.pcelm.append(data)

    def precomp_assignments(self):
        prefix = self.iq_prefix()
        code = []
        for data in self.pcelm:
            decl, er, ngrads, r = data
            for i in range(er.local_dimension):
                for j in range(er.value_size): # FIXME: symmetries? other?
                    if ngrads == 0:
                        lhscode = (prefix, ArrayAccess(decl, (i,j)))
                        rhscode = "FIXME" #er.basis_function(...)
                        code.append(format_code_structure(Assignment(lhscode, rhscode)))
                    else:
                        assert False # FIXME
                        lhscode = (prefix, ArrayAccess(decl, (i,j)))
                        rhscode = "FIXME" #er.basis_function_derivative(...)
                        code.append(format_code_structure(Assignment(lhscode, rhscode)))
        return code

    def precomp_declarations(self):
        code = []
        for data in self.pcelm:
            decl, er, ngrads, r = data
            code.append(format_code_structure(decl))
        return code

    def generate_basis_precomputations(self):
        code = []
        return code

    def generate_coefficient_declarations(self, coefficient, num_derivatives, restricted):
        code = ['']
        basename = "w%d" % coefficient.count()

        shape = coefficient.shape()
        symmetry = coefficient.element().symmetry()
        size = product(shape)
        nsd = self.geomrep.sfc_cell.nsd

        #er = self.FIXME

        restriction = None # FIXME

        if num_derivatives == 0:
            all_derivatives = [()]
        else:
            assert num_derivatives == 1 # FIXME
            all_derivatives = [(d,) for d in range(nsd)] # FIXME: Permutations of (d,)*num_der..

        for derivatives in all_derivatives:
            dstr = self.derivative_prefix(derivatives)
            cstr = "[%d]" % (size,) if size > 1 else ""
            decl = "double %s;" % (dstr + basename + cstr)
            code += [decl]

            # Iterate over components and flattened indices
            assert not symmetry # FIXME
            #comp2index, components = build_component_numbering(shape, symmetry)
            #for index, component in components:
            for component in compute_indices(shape):
                compindex = component_to_index(component, shape)

                terms = []
                nbf = 1 # er.nbf # FIXME
                for dofindex in range(nbf):
                    dofname = self.gen_coefficient_dof_name(coefficient, dofindex, restriction)
                    doffunc = "1.0/*FIXME*/" # er.N_dof_code[dofindex][compindex] # "pc[iq]." # FIXME: Add to element ref
                    if doffunc not in ("0", "0.0"):
                        terms.append("%s * %s" % (dofname, doffunc))
                assert terms
                expr = " + ".join(terms)

                name = self.gen_coefficient_name(coefficient, derivatives, component, restriction)
                assign = "%s = %s;" % (name, expr)
                code += [assign]

        return code

    def generate_argument_declarations(self, argument, num_derivatives, restricted):
        code = []
        basename = "v%d" % argument.count()

        shape = argument.shape()
        symmetry = argument.element().symmetry()
        size = product(shape)
        nsd = self.geomrep.sfc_cell.nsd

        restriction = None # FIXME

        if num_derivatives == 0:
            all_derivatives = [()]
        else:
            assert num_derivatives == 1 # FIXME
            all_derivatives = [(d,) for d in range(nsd)] # FIXME: Permutations of (d,)*num_der..

        for derivatives in all_derivatives:
            dstr = self.derivative_prefix(derivatives)
            cstr = "[%d]" % (size,) if size > 1 else ""
            decl = "double %s;" % (dstr + basename + cstr)
            code += [decl]

            # Iterate over components and flattened indices
            assert not symmetry # FIXME
            #comp2index, components = build_component_numbering(shape, symmetry)
            #for index, component in components:
            for component in compute_indices(shape):
                index = component_to_index(component, shape)

                name = self.gen_argument_name(argument, derivatives, component, restriction)
                dstr = self.derivative_prefix(derivatives)
                assert not restriction
                # FIXME: If derivatives, apply geometry mapping!
                element_id, element_index = 0, index # FIXME
                expr = self.iq_prefix() + dstr + "phi%d[%d]" % (element_id, element_index)
                #expr = "0.0" # FIXME

                assign = "%s = %s;" % (name, expr)
                code += [assign]
        code += ['']
        return code

    def gen_coefficient_dof_name(self, o, index, restriction):
        assert not restriction # FIXME: How does this map into w array?
        code = "w[%d][%d]" % (o.count(), index)
        return code

    def gen_constant_coefficient_name(self, o, component, restriction):
        assert not restriction # FIXME: How does this map into w array?
        assert not o.element().symmetry() # FIXME: Ignoring symmetries here!
        index = component_to_index(component, o.shape())
        code = self.gen_coefficient_dof_name(o, index, restriction)
        return code

    def iq_prefix(self):
        if self.on_facet:
            prefix = "pc[facet][iq]."
        else:
            prefix = "pc[iq]."
        return prefix

    def derivative_prefix(self, derivatives):
        if derivatives:
            dstr = "d%s_" % ("".join(map(str,sorted(derivatives))),)
        else:
            dstr = ""
        return dstr

    def component_prefix(self, shape, component):
        if shape:
            index = component_to_index(component, shape)
            cstr = "[%d]" % index
        else:
            cstr = ""
        return cstr

    def component_decl(self, shape):
        if shape:
            size = product(shape) # FIXME symmetry
            cstr = "[%d]" % size
        else:
            cstr = ""
        return cstr

    def gen_form_argument_decl_name(self, basename, o, derivatives, restriction):
        assert not restriction # FIXME
        dstr = self.derivative_prefix(derivatives)
        cstr = self.component_decl(o.shape())
        code = dstr + basename + str(o.count()) + cstr
        return code

    def gen_coefficient_decl(self, o, derivatives, restriction):
        return self.gen_form_argument_decl("w", o, derivatives, restriction)

    def gen_argument_decl(self, o, derivatives, restriction):
        return self.gen_form_argument_decl("v", o, derivatives, restriction)

    def gen_form_argument_name(self, basename, o, derivatives, component, restriction):
        assert not restriction # FIXME
        dstr = self.derivative_prefix(derivatives)
        cstr = self.component_prefix(o.shape(), component)
        code = dstr + basename + str(o.count()) + cstr
        return code

    def gen_coefficient_name(self, o, derivatives, component, restriction):
        return self.gen_form_argument_name("w", o, derivatives, component, restriction)

    def gen_argument_name(self, o, derivatives, component, restriction):
        return self.gen_form_argument_name("v", o, derivatives, component, restriction)
