
from sfc.representation.geometryrepresentation import GeometryRepresentation

class GeometryCG(object):
    def __init__(self, geomrep):
        self.geomrep = geomrep

    def generate_declarations(self,
                              cellvolume,
                              circumradius,
                              cellsurfacearea,
                              facetarea,
                              facetnormal):
        # FIXME: Skip what's not needed
        code = []
        code += ['', '// Vertex coordinates in more compact notation']
        code += self.gen_vx_code()

        code += ['', '// Offset for affine geometry mapping']
        code += self.gen_x0_code()

        code += ['', '// Jacobi of geometry mapping']
        code += self.gen_G_code()

        code += ['', '// Reference cell volume ratio']
        code += self.gen_detGtmp_code()
        code += self.gen_detG_code()
        code += self.gen_detG_sign_code()

        code += ['', '// Inverted Jacobi of geometry mapping']
        code += self.gen_Ginv_code()

        code += ['', '// Cell derived UFL quantities']
        if cellvolume:
            code += self.gen_cell_volume_code()
        if circumradius:
            code += self.gen_circumradius_code()
        if cellsurfacearea:
            code += self.gen_cell_surface_code()
        if facetarea:
            code += self.gen_facet_area_code()
        if facetnormal:
            code += self.gen_facet_normal_code()

        return code

    def generate_x_dependent_declarations(self, x, D, facet_D):
        gr = self.geomrep
        nsd = gr.sfc_cell.nsd

        code = []
        code += ['', '// Local coordinates']
        code += self.gen_xi_code()

        if x:
            code += ['', '// Global coordinates']
            code += self.gen_x_code()

        assert not (D and facet_D)
        if D or facet_D:
            code += ['', '// Integration point scaling factor']
        if D:
            code += self.gen_D_code()
        if facet_D:
            code += self.gen_facet_D_code()

        return code

    def gen_xi_code(self):
        code = []
        gr = self.geomrep
        nsd = gr.sfc_cell.nsd
        if 0:
            coords = ', '.join('qp[%d]' % i for i in range(nsd))
            code += ["const double xi[%d] = { %s };" % (nsd, coords)]
            assert gr.xi_sym[0].printc() == "xi[0]"
        else:
            code += ["const double %s = qp[%d];" % (gr.xi_sym[i].printc(), i)
                     for i in range(nsd)]
            assert gr.xi_sym[0].printc() == "x"
        return code

    def gen_x_names(self):
        gr = self.geomrep
        nsd = gr.sfc_cell.nsd
        code = [gr.x_sym[i].printc() for i in range(nsd)]
        return code
    def gen_x_code(self):
        gr = self.geomrep
        nsd = gr.sfc_cell.nsd
        code = []

        if 0:
            coords = ', '.join('qp[%d]' % i for i in range(nsd))
            code += ["const double x[%d] = { %s };" % (nsd, coords)]
            assert gr.x_sym[0].printc() == "x[0]"
        else:
            code += ["const double %s = %s;" % (gr.x_sym[i].printc(), gr.x_expr[i].printc())
                     for i in range(nsd)]
            assert gr.x_sym[0].printc() == "x0"
        return code

    def gen_vx_names(self):
        "Return names of vertex coordinate "
        sc = self.geomrep.sfc_cell
        d = sc.nsd
        num_vertices = sc.num_vertices
        return [gr.vx_sym[i][j].printc()
                for i in range(nv)
                for j in range(d)]
    def gen_vx_code(self):
        gr = self.geomrep
        sc = gr.sfc_cell
        d = sc.nsd
        nv = sc.num_vertices
        code = ["const double %s = %s;" % (gr.vx_sym[i][j].printc(),
                                           gr.vx_expr[i][j].printc())
                for i in range(nv)
                for j in range(d)]
        return code

    def gen_x0_names(self):
        d = self.geomrep.sfc_cell.nsd
        return [self.geomrep.x0_sym[i].printc() % i for i in range(d)]
    def gen_x0_code(self):
        # x0 just just set equal to vertex 0
        return []
    def xgen_x0_code(self):
        d = self.geomrep.sfc_cell.nsd
        expr = ", ".join(self.geomrep.x0_expr[i].printc() for i in range(d))
        code = "const double x0[%d] = { %s };" % (d, expr)
        return [code]

    def gen_G_names(self):
        d = self.geomrep.sfc_cell.nsd
        return [self.geomrep.G_sym[i,j].printc()
                for i in range(d)
                for j in range(d)]
    def gen_G_code(self):
        gr = self.geomrep
        d = gr.sfc_cell.nsd
        if 0:
            expr = ", ".join(gr.G_expr[i,j].printc()
                             for i in range(d)
                             for j in range(d))
            code = ["const double G[%d] = { %s };" % (d*d, expr)]
        else:
            code = ["const double %s = %s;" % (gr.G_sym[i,j].printc(),
                                               gr.G_expr[i,j].printc())
                    for i in range(d) for j in range(d)]
        return code

    def gen_Ginv_names(self):
        d = self.geomrep.sfc_cell.nsd
        return [self.geomrep.Ginv_sym[i,j].printc()
                for i in range(d)
                for j in range(d)]
    def gen_Ginv_code(self):
        gr = self.geomrep
        d = gr.sfc_cell.nsd
        if 0:
            expr = ", ".join(gr.Ginv_expr[i,j].printc()
                             for i in range(d)
                             for j in range(d))
            code = ["const double Ginv[%d] = { %s };" % (d*d, expr)]
        else:
            code = ["const double %s = %s;" % (gr.Ginv_sym[i,j].printc(),
                                               gr.Ginv_expr[i,j].printc())
                    for i in range(d) for j in range(d)]
        return code

    def gen_detG_names(self):
        return [self.geomrep.detG_sym.printc()]
    def gen_detG_code(self):
        expr = self.geomrep.detG_expr.printc()
        code = "const double detG = %s;" % expr
        return [code]

    def gen_detGtmp_names(self):
        return [self.geomrep.detGtmp_sym.printc()]
    def gen_detGtmp_code(self):
        expr = self.geomrep.detGtmp_expr.printc()
        code = "const double detGtmp = %s;" % expr
        return [code]

    def gen_detG_sign_names(self):
        return [self.geomrep.detG_sign_sym.printc()]
    def gen_detG_sign_code(self):
        expr = self.geomrep.detG_sign_expr.printc()
        code = "const double detG_sign = %s;" % expr
        return [code]

    def gen_facet_D_names(self):
        return ["facet_D"]
    def gen_facet_D_code(self):
        gr = self.geomrep
        code = []
        code += ["switch (facet)", "{"]
        for f in range(gr.sfc_cell.num_facets):
            code += ["case %d:" % f]
            code += ["%s = %s;" % (gr.facet_D_sym.printc(),
                                   gr.facet_D_expr[f].printc())]
        code += ["}"]
        # D = facet scaling times quadrature weight
        code += ["const double %s = %s * %s;" % (gr.D_sym.printc(),
                                                 gr.facet_D_sym.printc(),
                                                 gr.quad_weight_sym.printc())]
        return code

    def gen_D_names(self):
        return ["D"]
    def gen_D_code(self):
        gr = self.geomrep
        code = []
        # FIXME: Consistent quad_weight scaling! * gr.quad_weight_sym?
        code += ["const double %s = %s;" % (gr.D_sym.printc(), gr.D_expr.printc())]
        return code

    def gen_facet_normal_names(self):
        d = self.geomrep.sfc_cell.nsd
        return [self.geomrep.n_sym[0].printc() % i for i in range(d)]
    def gen_facet_normal_code(self):
        gr = self.geomrep
        d = gr.sfc_cell.nsd
        code = []
        assert self.geomrep.n_sym[0].printc() == "n0"
        code = []
        code += ["double %s;" % (', '.join('n%d'%i for i in range(d)),)]
        code += ["switch (facet)", "{"]
        for f in range(gr.sfc_cell.num_facets):
            code += ["case %d:" % f]
            # Mapping local facet normal to global cell orientation:
            for i in range(d):
                name = gr.n_tmp_sym[i].printc()
                expr = gr.n_tmp_expr[f][i].printc()
                code += ["    %s = %s;" % (name, expr)]
        code += ["}"]
        # Normalize global facet normal:
        #normscale = self.detG_sign_sym.printc(), # FIXME: scale by sign?
        normscale = "1.0"
        norm2 = ' + '.join((gr.n_sym[i].printc() + '*' + gr.n_sym[i].printc()) for i in range(d))
        code += ['const double nnorm = %s * sqrt(%s);' % (normscale, norm2)]
        for i in range(d):
            code += ['%s /= nnorm;' % (gr.n_sym[i].printc(),)]
        return [code]

    def gen_cell_volume_names(self):
        return ["K_vol"]
    def gen_cell_volume_code(self):
        gr = self.geomrep
        sc = gr.sfc_cell
        expr = (gr.detG_sym * sc.reference_volume).printc() # FIXME: Validate this
        code = "const double K_vol = %s;" % expr
        return [code]

    def gen_circumradius_names(self):
        return ["K_rad"]
    def gen_circumradius_code(self):
        expr = "FIXME"
        code = "const double K_rad = %s;" % expr
        return [code]

    def gen_cell_surface_names(self):
        return ["K_surf"]
    def gen_cell_surface_code(self):
        expr = "FIXME"
        code = "const double K_surf = %s;" % expr
        return [code]

    def gen_facet_area_names(self):
        return ["F_area"]
    def gen_facet_area_code(self):
        expr = "FIXME"
        code = "const double F_area = %s;" % expr
        return [code]
