# -*- coding: utf-8 -*-
"""
This module contains uflacs based code generation tools for the ufc::*_integral classes.
"""

# Copyright (C) 2012-2012 Martin Sandve Alnes and Simula Resarch Laboratory
#
# This file is part of SyFi.
#
# SyFi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# SyFi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SyFi. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2008-08-13
# Last changed: 2012-05-25

from sfc.common.output import sfc_debug, sfc_error, sfc_assert, sfc_warning
import ufc_utils
from sfc.common.utilities import indices_subset, unique
from sfc.codegeneration.support_code import (generate_sign_correction_support_code,
                                             gen_A_reset_block)
from sfc.quadrature import (gen_quadrature_rule_definition,
                            gen_quadrature_rule_definitions)

from sfc.codegeneration.integralcgbase import IntegralCGBase

try:
    import uflacs
except:
    uflacs = None
if uflacs is not None:
    import uflacs.backends.cpp2
    import uflacs.backends.cpp2.compiler
    import uflacs.codeutils.format_code_structure
    from uflacs.backends.cpp2.compiler import compile_expression_body
    from uflacs.codeutils.format_code_structure import format_code_structure
    from sfc.codegeneration.sfc_uflacs_formatter import SfcUflacsFormatter

class UflacsIntegralCG(IntegralCGBase):
    def __init__(self, itgrep):
        IntegralCGBase.__init__(self, itgrep)
        sfc_assert(uflacs is not None, "Couldn't import uflacs.")

        # Construct an SFC C++ expression formatter
        self.target_formatter = SfcUflacsFormatter(itgrep)

    def hincludes(self):
        return []

    def cincludes(self):
        return []

    def local_includes(self):
        return []

    def header_template(self):
        raise NotImplementedError("Missing implementation of header_template().")

    def impl_template(self):
        raise NotImplementedError("Missing implementation of impl_template().")

    def generate_support_code(self):
        code = []

        # Write quadrature rule if we have one:
        fr = self.itgrep.formrep
        # TODO: Place quadrature rules in a separate shared file?
        if self.itgrep._on_facet:
            if fr.facet_quad_rules is not None:
                code += [gen_quadrature_rule_definitions(fr.facet_quad_rules)]
        else:
            if fr.quad_rule is not None:
                code += [gen_quadrature_rule_definition(fr.quad_rule)]

        # Write sign correction code if needed
        code += [generate_sign_correction_support_code(self.itgrep.formrep)]

        # Declare struct for precomputed variables
        decls = self.target_formatter.facg.precomp_declarations()
        code += ["", "struct %s::precomputed" % (self.itgrep.classname,), "{"]
        code += ["    " + decl for decl in decls]
        code += ["};", ""]

        return '\n'.join(code)

    def gen_constructor(self):
        code = []
        code += ["pc = new precomputed[%d];" % self.itgrep.num_quadrature_points]
        code += [""]
        code += self.target_formatter.define_quadrature_for_loop()
        code += ["{"]
        code += self.target_formatter.define_quadrature_loop_vars()
        lines = ['']
        for assign in self.target_formatter.facg.precomp_assignments():
            lines.append("    " + assign)
        code.extend(lines)
        code += ["}"]
        return '\n'.join(code)

    def gen_constructor_arguments(self):
        return ""

    def gen_initializer_list(self):
        return ""

    def gen_destructor(self):
        code = []
        code += ["delete [] pc;"]
        return '\n'.join(code)

    def gen_members(self):
        code = ["",
                "struct precomputed;",
                "precomputed * pc;",
                ""]
        return '\n'.join(code)

    def xgen_tabulate_tensor(self):
        code = []
        code += ["int num_quadrature_points = 1;"]
        code += ["double quadrature_points[1][3] = { { 0.0, 0.0, 0.0 } };"]
        code += ["double quadrature_weights[1] = { 1.0 };"]
        code += ["tabulate_tensor(A, w, c,",
                 "                num_quadrature_points,",
                 "                quadrature_points,"
                 "                quadrature_weights);"]
        return code

    def gen_tabulate_tensor(self):
        # Get integrand expression
        integral, = self.itgrep.integrals # TODO: Handle multiple integrals with metadata
        expr = integral.integrand()
        #measure = integral.measure()
        #metadata = measure.metadata()

        # Initial UFC/SFC specific setup code
        header = gen_A_reset_block(len(self.itgrep.indices))

        # Compile body with generic compiler routine
        body = compile_expression_body(expr, self.target_formatter, integrate=True)

        code = format_code_structure([header,body])
        return code

    def gen_tabulate_tensor_quadrature(self):
        return 'throw std::runtime_error("Not implemented in uflacs mode.");'

class UflacsCellIntegralCG(UflacsIntegralCG):
    """
    /// Tabulate the tensor for the contribution from a local cell
    virtual void tabulate_tensor(double* A,
                                 const double * const * w,
                                 const cell& c) const = 0;

    /// Tabulate the tensor for the contribution from a local cell
    /// using the specified reference cell quadrature points/weights
    virtual void tabulate_tensor(double* A,
                                 const double * const * w,
                                 const cell& c,
                                 unsigned int num_quadrature_points,
                                 const double * const * quadrature_points,
                                 const double* quadrature_weights) const = 0;
    """
    def __init__(self, itgrep):
        UflacsIntegralCG.__init__(self, itgrep)

    def header_template(self):
        return ufc_utils.cell_integral_header

    def impl_template(self):
        return ufc_utils.cell_integral_implementation

class UflacsExteriorFacetIntegralCG(UflacsIntegralCG):
    """
    /// Tabulate the tensor for the contribution from a local exterior facet
    virtual void tabulate_tensor(double* A,
                                 const double * const * w,
                                 const cell& c,
                                 unsigned int facet) const = 0;

    /// Tabulate the tensor for the contribution from a local exterior facet
    /// using the specified reference cell quadrature points/weights
    virtual void tabulate_tensor(double* A,
                                 const double * const * w,
                                 const cell& c,
                                 unsigned int num_quadrature_points,
                                 const double * const * quadrature_points,
                                 const double* quadrature_weights) const = 0;
    """
    def __init__(self, itgrep):
        UflacsIntegralCG.__init__(self, itgrep)

    def header_template(self):
        return ufc_utils.exterior_facet_integral_header

    def impl_template(self):
        return ufc_utils.exterior_integral_implementation

    def gen_tabulate_tensor_quadrature(self):
        return 'throw std::runtime_error("This function is a bug in UFC. Not implemented in uflacs mode.");'

class UflacsInteriorFacetIntegralCG(UflacsIntegralCG):
    """
    /// Tabulate the tensor for the contribution from a local interior facet
    virtual void tabulate_tensor(double* A,
                                 const double * const * w,
                                 const cell& c0,
                                 const cell& c1,
                                 unsigned int facet0,
                                 unsigned int facet1) const = 0;

    /// Tabulate the tensor for the contribution from a local interior facet
    /// using the specified reference cell quadrature points/weights
    virtual void tabulate_tensor(double* A,
                                 const double * const * w,
                                 const cell& c,
                                 unsigned int num_quadrature_points,
                                 const double * const * quadrature_points,
                                 const double* quadrature_weights) const = 0;
    """
    def __init__(self, itgrep):
        UflacsIntegralCG.__init__(self, itgrep)

    def header_template(self):
        return ufc_utils.interior_facet_integral_header

    def impl_template(self):
        return ufc_utils.interior_integral_implementation

    def gen_tabulate_tensor_quadrature(self):
        return 'throw std::runtime_error("This function is a bug in UFC. Not implemented in uflacs mode.");'

# For dynamic importing of code generation strategy:
CellIntegralCG = UflacsCellIntegralCG
ExteriorFacetIntegralCG = UflacsExteriorFacetIntegralCG
InteriorFacetIntegralCG = UflacsInteriorFacetIntegralCG
