
from sfc.common.output import sfc_assert

from ufl.classes import (Terminal, Indexed, Grad, Restricted,
                         Coefficient, Argument,
                         SpatialCoordinate,
                         FacetNormal, FacetArea,
                         CellVolume, CellSurfaceArea, FacetArea, Circumradius)
from ufl.common import product, component_to_index
import uflacs
from uflacs.codeutils.target_formatter import (CppDefaultFormatter,
                                               analyse_terminalish)
from sfc.geometry.geometric_quantities import *
from sfc.codegeneration.geometrycg import GeometryCG
from sfc.codegeneration.formargumentcg import FormArgumentCG

class SfcUflacsFormatter(CppDefaultFormatter):
    """C++ formatter class for SFC code generation."""

    def __init__(self, itgrep, object_names=None):
        CppDefaultFormatter.__init__(self, object_names)
        self.itgrep = itgrep
        self.geomcg = GeometryCG(itgrep.formrep.geomrep)
        self.facg = FormArgumentCG(itgrep.formrep, on_facet=self.itgrep._on_facet)
        self.form_argument_mapping = itgrep.formrep.formdata.function_replace_map

        # Configuration options for statement formatting
        self._output_basename = "A"

    def update_terminals(self, terminals):
        CppDefaultFormatter.update_terminals(self, terminals)
        self.facg.precompute(terminals, self.required)

    # --- Functions for generating definitions and other statements

    def define_registers(self, num_registers):
        return ['// Intermediate variables for subexpressions',
                'double s[%d];' % num_registers]

    def define_piecewise_geometry(self):
        # Default to no quantities
        cellvolume = False
        circumradius = False
        cellsurfacearea = False
        facetarea = False
        facetnormal = False

        # Set the ones needed
        for o in self.terminals:
            (t, ngrads, r) = analyse_terminalish(o)
            if isinstance(t, FacetNormal):
                assert ngrads == 0
                facetnormal = True
            if isinstance(t, CellVolume):
                assert ngrads == 0
                cellvolume = True
            if isinstance(t, Circumradius):
                assert ngrads == 0
                circumradius = True
            if isinstance(t, CellSurfaceArea):
                assert ngrads == 0
                cellsurfacearea = True
            if isinstance(t, FacetArea):
                assert ngrads == 0
                facetarea = True

        # Outsource to the GeometryCG class
        return self.geomcg.generate_declarations(cellvolume=cellvolume,
                                                 circumradius=circumradius,
                                                 cellsurfacearea=cellsurfacearea,
                                                 facetarea=facetarea,
                                                 facetnormal=facetnormal)

    def define_piecewise_coefficients(self):
        # Coefficients in DG0 and R do not need to be computed,
        # we can just refer directly to w[count][component].
        return ""

    def xdefine_quadrature_for_loop(self):
        d = self.itgrep.formrep.formdata.cell.d
        points = ((0.0,)*d,)
        weigths = (1.0,)
        return CppDefaultFormatter.define_quadrature_for_loop(self, points=points,
                                                              weights=weigths)

    def define_quadrature_for_loop(self):
        code = []

        # TODO: define num_quadrature_points elsewhere so tabulate_tensor_quadrature can just work
        code += ["const int num_quadrature_points = %d;" % self.itgrep.num_quadrature_points]

        code += ["for (int iq = 0; iq < num_quadrature_points; ++iq)"]
        return code

    def define_quadrature_loop_vars(self):
        if self.itgrep._on_facet:
            code = []
            code += ["// Fetch quadrature point and weight on facet"]
            code += ["const double quad_weight = quadrature_weights[facet][iq];"]
            code += ["const double *qp = quadrature_points[facet][iq];"]
        else:
            code = []
            code += ["// Fetch quadrature point and weight in cell"]
            code += ["const double quad_weight = quadrature_weights[iq];"]
            code += ["const double *qp = quadrature_points[iq];"]
        return code

    def define_x_dependent_geometry(self):
        code = []

        # FIXME: finalize D/facet_D vs quad_weight scaling
        f = self.itgrep._on_facet
        code += self.geomcg.generate_x_dependent_declarations(x=True, D=not f, facet_D=f)

        # Compute x dependent geometry, if any?
        # A spatially varying geometry mapping could be computed here.
        return code

    def define_x_dependent_coefficients(self):
        code = []
        for o in self.terminals:
            (t, ngrads, r) = analyse_terminalish(o)

            c = self.form_argument_mapping.get(t, t)

            if isinstance(c, Coefficient) and not c.is_cellwise_constant():
                code += self.facg.generate_coefficient_declarations(c, ngrads, bool(r))
        return code

    def define_argument_for_loop(self, argument_count):
        iname = "i%d" % (argument_count,)

        fr = self.itgrep.formrep
        dim = fr.element_reps[fr.formdata.elements[argument_count]].local_dimension
        isize = str(dim)

        code = ["for (int %s = 0; %s < %s; ++%s)" % (iname, iname, isize, iname)]
        return code

    def define_argument_loop_vars(self, argument_count):
        # Generate code for all needed derivatives of arguments with this count
        code = []
        handled = set()
        for o in self.terminals:
            (t, ngrads, r) = analyse_terminalish(o)

            c = self.form_argument_mapping.get(t, t)

            if isinstance(c, Argument) and c.count() == argument_count:
                # Avoid duplicates (restrictions will occur with + and -)
                key = (c, ngrads, bool(r))
                if key not in handled:
                    handled.add(key)
                    code += self.facg.generate_argument_declarations(c, ngrads, bool(r))
        return code

    def output_variable_names(self, num_variables):
        return ['%s[%d]' % (self._output_basename, i,) for i in xrange(num_variables)]

    # --- Explicitly selecting rules for geometry here just for clarity:
    def spatial_coordinate(self, o, component=(), derivatives=(), restriction=None):
        # Restriction has no effect on x
        #postfix = {None:"", "+": "_p", "-": "_m"}[restriction]

        sfc_assert(not derivatives, "I don't think this should happen.")
        if derivatives:
            sfc_warning("Compiler should be able to simplify derivatives of x!")
            code = "1" if (len(derivatives) == 1 and i == derivatives[0]) else "0"
        else:
            i = component[0] if component else 0
            code = self.geomcg.gen_x_names()[i]
            self.require(o, component, derivatives, restriction, code)
            return code

    # n[0] etc.
    #facet_normal = CppDefaultFormatter.facet_normal
    def facet_normal(self, o, component=(), derivatives=(), restriction=None):
        sfc_assert(not derivatives, "I don't think this should happen.")
        i = component[0] if component else 0

        #postfix = {None:"", "+": "_p", "-": "_m"}[restriction]
        code = self.geomcg.gen_facet_normal_names()[i]

        self.require(o, component, derivatives, restriction, code)
        return code

    # K_vol
    cell_volume = CppDefaultFormatter.cell_volume

    # K_rad
    circumradius = CppDefaultFormatter.circumradius

    # K_surf
    cell_surface_area = CppDefaultFormatter.cell_surface_area

    # F_area
    facet_area = CppDefaultFormatter.facet_area

    # --- Rules for coefficients and arguments:

    def coefficient(self, o, component=(), derivatives=(), restriction=None):
        #basename = self.coefficient_names[o]
        o = self.form_argument_mapping.get(o, o)

        if o.is_cellwise_constant():
            # For piecewise constant functions, refer directly to the dof symbol w[][]
            assert not derivatives
            code = self.facg.gen_constant_coefficient_name(o, component, restriction)
        else:
            code = self.facg.gen_coefficient_name(o, derivatives, component, restriction)

        self.require(o, component, derivatives, restriction, code)
        return code

    def argument(self, o, component=(), derivatives=(), restriction=None):
        #basename = self.argument_names[o]
        o = self.form_argument_mapping.get(o, o)
        code = self.facg.gen_argument_name(o, derivatives, component, restriction)
        self.require(o, component, derivatives, restriction, code)
        return code
