#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module contains the Just-In-Time compiler tools.
The main function is jit(form, options), see its documentation.
"""

# Copyright (C) 2008-2009 Martin Sandve Alnes and Simula Resarch Laboratory
#
# This file is part of SyFi.
#
# SyFi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# SyFi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SyFi. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Kent-Andre Mardal, 2010.
#
# First added:  2008-09-04
# Last changed: 2009-04-23

import shutil
import os
import hashlib
import instant
import ufc_utils
import ufl
import ufl.algorithms

from sfc.common import version
from sfc.common import default_parameters
from sfc.common import ParameterDict
from sfc.common import get_abs_ufc_h_path
from sfc.common import sfc_debug, sfc_info, sfc_warning, sfc_error, sfc_assert, write_file
from sfc.common.names import short_name, form_classname, dof_map_classname, finite_element_classname, base_element_classname
from sfc.codegeneration import generate_code, compiler_input

def cache_signature(ufl_elements, formdatas, parameters, prefix):
    "A signature string uniquely identifying jit input."
    h = hashlib.md5(repr(version))
    h.update(prefix)
    for e in ufl_elements:
        h.update(repr(e))
    for fd in formdatas:
        h.update(fd.signature)
    h.update(repr(parameters))
    return h.hexdigest().lower()

def compile_module(module_dir, hfilenames, cfilenames, signature, options):
    """Assuming an existing directory module_dir with source files
    hfilenames, cfilenames, create a python extension module and compile it."""

    orig_dir = os.getcwd()
    try:
        # headers to #include in swig file
        system_headers = ["iostream", "stdexcept", "cmath", "ufc.h", "tr1/memory"]

        cppargs = options.compilation.cppargs
        if options.compilation.enable_debug_code:
            cppargs.append("-DSFCDEBUG")

        module_name = module_dir if signature is None else None

        module = ufc_utils.build_ufc_module(h_files = hfilenames,
                     source_directory   = module_dir,
                     sources            = cfilenames,
                     system_headers     = system_headers,
                     cppargs            = cppargs,
                     signature          = signature,
                     cache_dir          = options.compilation.cache_dir,
                     modulename         = module_name,
                     generate_interface = options.compilation.generate_interface,
                     generate_setup     = options.compilation.generate_setup)
        sfc_info("Successfully compiled module in '%s'." % module_dir)
        return module
        # ...
    finally:
        os.chdir(orig_dir)

def merge_with_default_parameters(parameters):
    if parameters is None:
        parameters = default_parameters()
    if isinstance(parameters, dict) and not isinstance(parameters, ParameterDict):
        o = parameters
        parameters = default_parameters()
        parameters.update(o)
    return parameters

def lookup_module(parameters, signature):
    module = None
    if parameters.compilation.skip:
        # Hacky hook for profiling without the C++ compilation
        sfc_warning("Skipping module lookup.")

    elif parameters.compilation.overwrite_cache:
        # Possibly force recompilation
        sfc_warning("Complete implementation of overwrite_cache not done, but skipping cache lookup before code generation.")
        # TODO: delete existing module with signature or module_name

    else:
        # Check cache before generating code
        module = instant.import_module(signature)
        # Construct ufc::form object
        if module:
            sfc_info("Found module in cache.")

    return module

def make_module_dir(prefix, formclassnames, dmclassnames, feclassnames):
    # Build module name, then shorten if it's too long
    module_dir_name = "_".join(["_".join(formclassnames),
                                "_".join(dmclassnames),
                                "_".join(feclassnames)])
    module_dir_name = short_name(prefix, module_dir_name)

    # Make sure the module directory is there and empty
    module_dir = os.path.abspath(module_dir_name)
    if os.path.exists(module_dir):
        msg = "Possibly overwriting files in existing module directory '%s'."
        sfc_warning(msg % module_dir)
    else:
        os.mkdir(module_dir)
    return module_dir

def write_parameters(module_dir, parameters, signature, ufl_elements):
    # Write parameters, signature and form representations to files beside code
    write_file(os.path.join(module_dir, "parameters.repr"), repr(parameters))
    write_file(os.path.join(module_dir, "signature"), signature)
    for ue in ufl_elements:
        sn = short_name("element_", base_element_classname(ue))
        write_file(os.path.join(module_dir,"%s.repr" % sn), repr(ue))

def pack_jit_return_values(formdatas, ufl_elements, module,
                           formclassnames, feclassnames, dmclassnames,
                           list_input):
    # Got forms, return forms
    if formdatas:
        prefix = "sfc dummy prefix for pydolfin which I have no idea what is used for"
        ufc_forms = [getattr(module, name)() for name in formclassnames]
        if list_input:
            return ufc_forms, module, formdatas, [prefix]*len(ufc_forms)
        else:
            return ufc_forms[0], module, formdatas[0], prefix

    # Got elements, return elements
    elif ufl_elements:
        ufc_elements = [(getattr(module, fe)(), getattr(module, dm)())
                        for (fe, dm) in zip(feclassnames, dmclassnames)]
        if list_input:
            return ufc_elements
        else:
            return ufc_elements[0]

    msg = "Nothing to return, this shouldn't happen.\ninput =\n%r" % repr(input)
    sfc_error(msg)

def jit(input, parameters=None, common_cell=None, objects=None):
    # To handle all input uniformly
    list_input = isinstance(input, list)
    ufl_elements, formdatas = compiler_input(input, objects=objects,
                                             common_cell=common_cell)
    parameters = merge_with_default_parameters(parameters)

    # Default prefix shows where the module was compiled
    prefix = parameters.code.prefix or "sfc_jit"

    # Build module sig from input data on canonical form
    signature = cache_signature(ufl_elements, formdatas, parameters, prefix)

    # Build classnames (TODO: This is duplicate work, done in formrepresentation)
    formclassnames = [form_classname(fd.signature, parameters)
                      for fd in formdatas]
    dmclassnames = [dof_map_classname(e) for e in ufl_elements]
    feclassnames = [finite_element_classname(e) for e in ufl_elements]

    # Lookup module in cache before compiling
    module = lookup_module(parameters, signature)
    if module is None:
        module_dir = make_module_dir(prefix, formclassnames, dmclassnames, feclassnames)

        write_parameters(module_dir, parameters, signature, ufl_elements)

        # Enter module directory and generate code!
        orig_dir = os.getcwd()
        try:
            os.chdir(module_dir)
            hfilenames, cfilenames = generate_code(ufl_elements + formdatas, objects,
                                                   parameters, common_cell=common_cell)
        finally:
            # Restore original path
            os.chdir(orig_dir)

        # Hook for profiling without the C++ compilation
        if parameters.compilation.skip:
            return None

        # Invoke Instant to compile generated files as a Python extension module
        module = compile_module(module_dir, hfilenames, cfilenames,
                                signature, parameters)

        # Remove temporary module directory if code was copied to cache
        placed_in_cache = False # TODO
        if placed_in_cache:
            msg = "Module was placed in cache, deleting module directory '%s'."
            sfc_info(msg % module_dir)
            shutil.rmtree(module_dir)

        if module is None:
            sfc_error("Failed to load compiled module from cache.")

    # If we get here, 'module' should be our compiled module.

    # TODO: Handle mixed list of forms and elements?

    return pack_jit_return_values(formdatas, ufl_elements, module,
                                  formclassnames, feclassnames, dmclassnames,
                                  list_input)

# TODO: Get rid of this one
def jitf(input, objects, parameters=None, common_cell=None):
    return jit(input, parameters=parameters, common_cell=common_cell, objects=objects)
