# -*- coding: utf-8 -*-
"""
This module contains a representation class for forms, storing information about all form aspects.
"""

# Copyright (C) 2008-2009 Martin Sandve Alnes and Simula Resarch Laboratory
#
# This file is part of SyFi.
#
# SyFi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# SyFi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SyFi. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2008-08-13
# Last changed: 2009-04-19

import copy
import math

import swiginac
import SyFi

import ufl
from ufl.algorithms import extract_max_quadrature_element_degree,\
    estimate_quadrature_degree, estimate_total_polynomial_degree

from sfc.common import sfc_assert
from sfc.common.names import finite_element_classname, dof_map_classname, \
                           integral_classname, form_classname
from sfc.common.utilities import (matrix_to_list, list_to_vector,
                                  index_string, dot_product,
                                  build_count_dict)
from sfc.geometry import geometry_mapping
from sfc.quadrature import find_quadrature_rule
from sfc.symbolic_utils import symbolic_matrix, symbolic_vector, symbol, symbols, cross, sub, inner
from sfc.representation.elementrepresentation import create_syfi_polygon
from sfc.representation.geometryrepresentation import GeometryRepresentation

class FormArgumentRepresentation(object):
    def __init__(self, formrep):
        self.formrep = formrep

        # Storing some references here for convenience.
        # (Note that this class stems from a partial refactoring of old FormRepresentation class)
        self.formdata = formrep.formdata
        self.element_reps = formrep.element_reps
        self.geomrep = formrep.geomrep
        self.rank = formrep.rank
        self.num_coefficients = formrep.num_coefficients

        # ... Build symbols for all coefficient dofs as UFC input variables w[][]
        # Get dimensions from element representations in coefficient ordering
        dims = [self.element_reps[self.formdata.elements[self.rank+i]].local_dimension
                for i in range(self.num_coefficients)]
        # Build names
        self.w_dof_names = []
        for i, dim in enumerate(dims):
            wi_dof_names = ["w[%d][%d]" % (i, j) for j in range(dim)]
            self.w_dof_names.append(wi_dof_names)
        # Build symbols
        self.w_dofs = [symbols(wdn) for wdn in self.w_dof_names]

        # ... TODO: Build dofindex-templated names for all references to precomputed basis functions?

    def __str__(self):
        s = ""
        s += "    Argument tokens:\n"
        s += "        self.w_dofs          = %s\n" % self.w_dofs
        return s

    # --- Basis function and coefficient data for arguments

    def v_expr(self, iarg, i, component):
        elm = self.formdata.elements[iarg]
        #component, elm = elm.extract_component(component)

        rep = self.element_reps[elm]
        v = rep.basis_function(i, component)
        return v

    def v_sym(self, iarg, i, component, on_facet):
        elm = self.formdata.elements[iarg]

        e = self.v_expr(iarg, i, component)
        e2 = e.evalf()
        if e2.nops() == 0:
            # FIXME: Remove this line after fixing code generation,
            #        this leads to "double 0;" statements:
            if e2 == 0:
                return e2

        scomponent, selm = elm.extract_component(component)
        num = self.formrep.element_number[selm]

        prefix = "qt[facet][iq]." if on_facet else "qt[iq]."
        suffix = index_string((num, i) + scomponent)
        name = "%sv_%s" % (prefix, suffix)
        s = symbol(name)
        return s

    def w_expr(self, iarg, component, use_symbols, on_facet):
        elm = self.formdata.elements[self.rank+iarg]
        rep = self.element_reps[elm]
        dim = rep.local_dimension
        if use_symbols:
            vs = [self.v_sym(iarg+self.rank, i, component, on_facet=on_facet)
                  for i in range(dim)]
        else:
            vs = [self.v_expr(iarg+self.rank, i, component) for i in range(dim)]
        w = dot_product(self.w_dofs[iarg], vs)
        return w

    def w_sym(self, iarg, component):
        name = "w_%s" % index_string((iarg,) + component)
        s = symbol(name)
        return s

    # --- Basis function and coefficient data for derivatives of arguments

    def ddxi(self, f, i):
        gr = self.formrep.geomrep
        return swiginac.diff(f, gr.xi_sym[i])

    def ddx(self, f, i):
        gr = self.formrep.geomrep
        # TODO: Verify this line (in particular transposed or not!)
        return sum(gr.Ginv_sym[j,i]*self.ddxi(f, j)
                   for j in range(gr.sfc_cell.nsd))

    def dv_expr(self, iarg, i, component, d):
        """Return expression for dv/dxi, with v being a particular
        component of basis function i in argument space iarg,
        and d is a tuple (i.e. multiindex) of xi directions (local coordinates)."""
        sfc_assert(len(d) == 1, "Higher order derivatives not implemented.")
        d = tuple(sorted(d))
        v = self.v_expr(iarg, i, component)
        dv = v
        for k in d:
            dv = self.ddxi(dv, k)
        return dv

    def dv_sym(self, iarg, i, component, d, on_facet):
        sfc_assert(len(d) == 1, "Higher order derivatives not implemented.")
        d = tuple(sorted(d))

        e = self.dv_expr(iarg, i, component, d)
        e2 = e.evalf()
        if e2.nops() == 0:
            # FIXME: Remove this line after fixing code generation,
            #        this leads to "double 0;" statements:
            if e2 == 0:
                return e2

        prefix = "qt[facet][iq]." if on_facet else "qt[iq]."
        suffix = index_string((iarg, i) + component + d)
        name = "%sdv_dxi_%s" % (prefix, suffix)
        return symbol(name)

    def Dv_expr(self, iarg, i, component, d, use_symbols, on_facet):
        """Return expression for dv/dx, with v being a particular
        component of basis function i in argument space iarg,
        and d is a tuple (i.e. multiindex) of x directions (global coordinates)."""
        sfc_assert(len(d) == 1, "Higher order derivatives not implemented.")
        d = tuple(sorted(d))
        # Get local derivatives in all directions
        gr = self.formrep.geomrep
        if use_symbols:
            dv = [self.dv_sym(iarg, i, component, (j,), on_facet=on_facet)
                  for j in range(gr.sfc_cell.nsd)]
        else:
            dv = [self.dv_expr(iarg, i, component, (j,))
                  for j in range(gr.sfc_cell.nsd)]
        # Apply mapping to dv
        j, = d
        # TODO: Verify this line (in particular transposed or not!)
        Dv = sum(gr.Ginv_sym[k, j]*dv[k] for k in range(gr.sfc_cell.nsd))
        return Dv

    def Dv_sym(self, iarg, i, component, d, on_facet):
        sfc_assert(len(d) == 1, "Higher order derivatives not implemented.")
        d = tuple(sorted(d))

        e = self.Dv_expr(iarg, i, component, d, False, on_facet=on_facet)
        e2 = e.evalf()
        if e2.nops() == 0:
            # FIXME: Remove this line after fixing code generation,
            #        this leads to "double 0;" statements:
            if e2 == 0:
                return e2

        name = "Dv_%s" % index_string((iarg, i) + component + d)
        return symbol(name)

    def Dw_expr(self, iarg, component, d, use_symbols, on_facet):
        sfc_assert(len(d) == 1, "Higher order derivatives not implemented.")
        d = tuple(sorted(d))
        elm = self.formdata.elements[self.rank+iarg]
        rep = self.formrep.element_reps[elm]
        dim = rep.local_dimension
        if use_symbols:
            vs = [self.Dv_sym(self.rank+iarg, i, component, d, on_facet=on_facet)
                  for i in range(dim)]
        else:
            vs = [self.Dv_expr(self.rank+iarg, i, component, d,
                               use_symbols=False, on_facet=on_facet)
                  for i in range(dim)]
        w = dot_product(self.w_dofs[iarg], vs)
        return w

    def Dw_sym(self, iarg, component, d):
        sfc_assert(len(d) == 1, "Higher order derivatives not implemented.")
        d = tuple(sorted(d))
        name = "Dw_%s" % index_string((iarg,) + component + d)
        return symbol(name)
