# -*- coding: utf-8 -*-
"""
This module contains a representation class for forms, storing information about all form aspects.
"""

# Copyright (C) 2008-2009 Martin Sandve Alnes and Simula Resarch Laboratory
#
# This file is part of SyFi.
#
# SyFi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# SyFi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SyFi. If not, see <http://www.gnu.org/licenses/>.
#
# First added:  2008-08-13
# Last changed: 2009-04-19

import copy
import math

import swiginac
import SyFi

import ufl
from ufl.algorithms import extract_max_quadrature_element_degree,\
    estimate_quadrature_degree, estimate_total_polynomial_degree

from sfc.common import sfc_assert
from sfc.common.names import finite_element_classname, dof_map_classname, \
                           integral_classname, form_classname
from sfc.common.utilities import (matrix_to_list, list_to_vector,
                                  index_string, dot_product,
                                  build_count_dict)
from sfc.geometry import geometry_mapping
from sfc.quadrature import find_quadrature_rule
from sfc.symbolic_utils import symbolic_matrix, symbolic_vector, symbol, symbols, cross, sub, inner
from sfc.representation.elementrepresentation import create_syfi_polygon
from sfc.representation.geometryrepresentation import GeometryRepresentation
from sfc.representation.formargumentrepresentation import FormArgumentRepresentation

# FIXME: Need some refactoring of this class,
# for example splitting out geometry and function code in separate classes,
# and splitting out naming from symbol creation.
class FormRepresentationBase(object):
    def __init__(self, formdata, element_reps, geomrep, options):

        # UFL data about the form
        self.formdata = formdata

        # Mapping : ufl.FiniteElement -> sfc.ElementRepresentation
        self.element_reps = element_reps

        # Shared geometry representation data
        self.geomrep = geomrep

        # SFC options
        self.options   = options

        # Fetch some common variables from formdata for shorter names in rest of code
        self.form             = self.formdata.preprocessed_form
        self.rank             = self.formdata.rank
        self.num_coefficients = self.formdata.num_coefficients

        self.polygon_name     = self.formdata.cell.domain()
        self.domain = self.polygon_name # TODO: Rename everywhere and remove

        self.signature = self.formdata.signature

        # Some strings
        self._define_classnames()

        # Build running count of all _unique_ ufl elements.
        # Not the same as running count of form arguments!
        # mapping : ufl_element -> running count of unique elements
        self.element_number = build_count_dict(self.formdata.sub_elements)

    def _define_classnames(self):
        # Generate names for all classes
        f = self.form
        fd = self.formdata

        self.classname = form_classname(fd.signature, self.options)
        self.fe_names  = [finite_element_classname(e) for e in fd.elements]
        self.dm_names  = [dof_map_classname(e) for e in fd.elements]

        self.itg_names = dict((itg, integral_classname(itg, self.classname)) \
                              for itg in self.form.integrals())

        self.citg_names = dict((itg.measure().domain_id(), self.itg_names[itg])
                               for itg in f.cell_integrals())
        self.eitg_names = dict((itg.measure().domain_id(), self.itg_names[itg])
                               for itg in f.exterior_facet_integrals())
        self.iitg_names = dict((itg.measure().domain_id(), self.itg_names[itg])
                               for itg in f.interior_facet_integrals())

class QuadratureRuleRepresentation(object):
    def __init__(self, form, sfc_cell, options):
        # Determine quadrature rule from options or elements
        self.quad_order = decide_quad_order(options, form)

        # Pick a rule for cell
        self.quad_rule = find_quadrature_rule(sfc_cell.shape, self.quad_order)

        # Pick a rule for reference facet
        reference_quad_rule = find_quadrature_rule(sfc_cell.facet_shape,
                                                   self.quad_order, family="gauss")
        self.facet_quad_rules = map_facet_quadrature_rules(reference_quad_rule, sfc_cell)

def decide_quad_order(options, form):
    # Get integration order from options or elements
    d = options.code.integral.integration_order

    if d is None:
        d = extract_max_quadrature_element_degree(form)
    if d is None:
        # TODO: Option for which estimation method to choose!
        d = estimate_quadrature_degree(form)
    if d is None:
        d = estimate_total_polynomial_degree(form)

    sfc_assert(d is not None, "All quadrature degree estimation rules failed!")
    return d

def map_facet_quadrature_rules(reference_quad_rule, sfc_cell):
    # Map rule to each facet polygon on the reference cell
    facet_quad_rules = []
    for facet, facet_polygon in enumerate(sfc_cell.facet_polygons):

        # TODO: That dimensions get right here is rather magical,
        #       the global dimension set by initSyFi plays a major role here...

        # affine mapping between reference
        facet_G, facet_x0 = geometry_mapping(facet_polygon)

        # scaling factor for diagonal facets,
        # for scaling quadrature rule to a
        # different reference domain, this is NOT the
        # same as the global/local scaling factor facet_D!
        D = 1.0
        if facet == 0:
            if sfc_cell.shape == "triangle":
                D = math.sqrt(2)
            elif sfc_cell.shape == "tetrahedron":
                D = math.sqrt(3)

        # scale weights by scaling factor
        weights = [float(w)*D for w in reference_quad_rule.weights]

        # apply affine mapping to quadrature points
        points = []
        for p in reference_quad_rule.points:
            # extend (n-1)D point p into nD point x:
            x = swiginac.matrix(len(facet_x0), 1, list(p)+[0.0])
            # apply mapping in nD space:
            newx = facet_G * x + facet_x0
            # store as a list:
            xlist = [float(newx[i].evalf()) for i in range(len(newx))]
            points.append(xlist)

        # make a copy of reference quadrature rule and modify it:
        facet_quad_rule = copy.copy(reference_quad_rule)

        if sfc_cell.shape != "interval":
            facet_quad_rule.nsd += 1
            assert facet_quad_rule.nsd == sfc_cell.nsd
        facet_quad_rule.comment += "\nMapped to polygon in higher dimension.\n"
        facet_quad_rule.weights = weights
        facet_quad_rule.points = points
        facet_quad_rules.append(facet_quad_rule)
    return facet_quad_rules

# FIXME: Access subclasses through members instead, this is a partial refactoring state.
class FormRepresentation(FormRepresentationBase,
                         GeometryRepresentation,
                         QuadratureRuleRepresentation,
                         FormArgumentRepresentation):
    def __init__(self, formdata, element_reps, geomrep, options):
        if 0:
            quadrep = QuadratureRuleRepresentation(formdata.preprocessed_form,
                                                   geomrep.sfc_cell, options)
            farep = FormArgumentRepresentation(self)

        FormRepresentationBase.__init__(self, formdata, element_reps, geomrep, options)

        if 1:
            GeometryRepresentation.__init__(self, formdata.cell.domain())
            QuadratureRuleRepresentation.__init__(self, formdata.preprocessed_form,
                                                  geomrep.sfc_cell, options)
            FormArgumentRepresentation.__init__(self, self)

    def __str__(self):
        s = ""
        s += "FormRepresentation:\n"
        s += "\n" # TODO: Add more here
        s += GeometryRepresentation.__str__(self)
        s += FormArgumentRepresentation.__str__(self)
        return s

    #def element_representation(self, iarg):
    #    return self.element_reps[self.formdata.elements[iarg]]

    #def unique_element_number(self, iarg):
    #    return self.element_number[self.formdata.elements[iarg]]

    #===============================================================================
    #    def _build_argument_cache(self):
    #        use_symbols = self.integration_method == "quadrature"
    #
    #        self.v_cache = []
    #        self.Dv_cache = []
    #        for iarg in range(self.rank + self.num_coefficients):
    #            elm = self.formdata.elements[iarg]
    #            rep = self.element_reps[elm]
    #            self.v_cache.append([])
    #            self.Dv_cache.append([])
    #            for i in range(rep.local_dimension):
    #                for component in rep.value_components:
    #                    ve = self.v_expr(iarg, i, component)
    #                    vs = self.v_sym(iarg, i, component)
    #                    self.v_cache[iarg][i][component] = (vs, ve)
    #
    #                    for derivatives in [(d,) for d in range(self.sfc_cell.nsd)]:
    #                        Dve = self.Dv_expr(iarg, i, component, derivatives, use_symbols)
    #                        Dvs = self.Dv_sym(iarg, i, component, derivatives)
    #                        self.Dv_cache[iarg][i][component][derivatives] = (Dvs, Dve)
    #===============================================================================

