
import swiginac
import SyFi

from sfc.symbolic_utils import symbolic_matrix, symbolic_vector, symbol, symbols, cross, sub, inner
from sfc.common.utilities import matrix_to_list, list_to_vector, index_string, dot_product
from sfc.geometry import SFCCell, geometry_mapping

# TODO: Replace nsd with topological and geometric dimensions
# everywhere for clarity and future flexibility

def create_syfi_polygon(cell):
    if cell == "interval":
        return SyFi.ReferenceLine()
    elif cell == "triangle":
        return SyFi.ReferenceTriangle()
    elif cell == "tetrahedron":
        return SyFi.ReferenceTetrahedron()
    elif cell == "quadrilateral":
        return SyFi.ReferenceRectangle()
    elif cell == "hexahedron":
        return SyFi.ReferenceBox()
    raise ValueError("Unknown element cell '%s'." % cell)


def build_global_polygon(shape, p):
    if shape == "interval":
        polygon = SyFi.Line(*p)
    elif shape == "triangle":
        polygon = SyFi.Triangle(*p)
    elif shape == "tetrahedron":
        polygon = SyFi.Tetrahedron(*p)
    elif shape == "quadrilateral":
        #SyFi.Triangle(p[0], p[1], p[3])
        #SyFi.Rectangle(p[0], p[2])     # TODO: Better affine mapping?
        polygon = SyFi.Rectangle(*p)
    elif shape == "hexahedron":
        # SyFi.Tetrahedron(p[0], p[1], p[3], p[4])
        # SyFi.Box(p[0], p[6]) # TODO: Better affine mapping?
        polygon = SyFi.Box(*p)
    return polygon


class GeometryRepresentation(object):
    def __init__(self, ufl_cell_name):
        self.polygon_name = ufl_cell_name
        self._build_geometry_tokens()

    def __str__(self):
        s = ""
        s += "    Geometry tokens:\n"
        s += "        self.sfc_cell        = %s\n" % self.sfc_cell
        #s += "        self.global_cell     = %s\n" % self.global_cell
        s += "        self.vx_sym          = %s\n" % self.vx_sym
        s += "        self.vx_expr         = %s\n" % self.vx_expr
        s += "        self.xi_sym          = %s\n" % self.xi_sym
        s += "        self.x_sym           = %s\n" % self.x_sym
        s += "        self.x_expr          = %s\n" % self.x_expr
        s += "        self.G_sym           = %s\n" % self.G_sym
        s += "        self.G_expr,         = %s\n" % self.G_expr,
        s += "        self.detGtmp_sym     = %s\n" % self.detGtmp_sym
        s += "        self.detGtmp_expr    = %s\n" % self.detGtmp_expr
        s += "        self.detG_sym        = %s\n" % self.detG_sym
        s += "        self.detG_expr       = %s\n" % self.detG_expr
        s += "        self.Ginv_sym        = %s\n" % self.Ginv_sym
        s += "        self.Ginv_expr       = %s\n" % self.Ginv_expr
        s += "        self.quad_weight_sym = %s\n" % self.quad_weight_sym
        return s

    def _build_geometry_tokens(self):
        "Build tokens for variables derived from cell."

        # FIXME: Use unambiguous names: ufl_cell, sfc_cell, syfi_polygon, global_syfi_polygon
        self.syfi_polygon = create_syfi_polygon(self.polygon_name)
        self.sfc_cell = SFCCell(self.syfi_polygon)

        self.cell = self.sfc_cell # FIXME: Rename all references to this member to sfc_cell

        sfc_cell = self.sfc_cell
        nsd = sfc_cell.nsd

        # vx[i][j] = component j of coordinate of vertex i in cell
        self.vx_sym = []
        self.vx_expr = []
        for i in range(sfc_cell.num_vertices):
            s = swiginac.matrix(nsd, 1, symbols("vx%d_%d" % (i,j) for j in range(nsd)))
            e = swiginac.matrix(nsd, 1, symbols("c.coordinates[%d][%d]" % (i,j)
                                                for j in range(nsd)))
            self.vx_sym.append(s)
            self.vx_expr.append(e)

        # Build a global polygon from the vertex coordinates:
        #   (using Triangle and Tetrahedron for the hypercube affine mappings
        #   is correct for rectangular and trapezoidal shapes, while using the polygons
        #   Box and Rectangle from SyFi would only support straight rectangular shapes)
        p = [matrix_to_list(vx) for vx in self.vx_sym]
        self.global_polygon = build_global_polygon(sfc_cell.shape, p)

        # Geometry mapping
        # FIXME: Make sure we don't mess up transpose here again:
        self.G_sym  = symbolic_matrix(nsd, nsd, "G")
        # FIXME: Make sure this is consistent:
        self.x0_sym = self.vx_sym[0] #symbolic_vector(nsd, "x0")
        # TODO: Non-affine mappings:
        self.G_expr, self.x0_expr = geometry_mapping(self.global_polygon)

        # FIXME: Rename so that detG includes the sign and detGabs is the absolute, detGtmp/detG is unclear
        self.detGtmp_sym  = symbol("detGtmp")
        self.detGtmp_expr = swiginac.determinant(self.G_sym)
        self.detG_sym  = symbol("detG")
        #self.detG_expr = self.detGtmp_sym
        self.detG_expr = swiginac.abs(self.detGtmp_sym)

        # Sign of det G, negative if cell is inverted
        self.detG_sign_sym = symbol("detG_sign")
        self.detG_sign_expr = self.detG_sym / swiginac.abs(self.detG_sym)

        # Inverse of geometry mapping (expression simplification using cofactor)
        self.Ginv_sym  = symbolic_matrix(nsd, nsd, "Ginv")
        self.Ginv_expr = ((self.detGtmp_expr * swiginac.inverse(self.G_sym))
                          / self.detGtmp_sym)
        #self.Ginv_expr = swiginac.inverse(self.G_sym)

        # Local and global coordinates (getting local xi names from cell)
        # TODO: Change in cell to "xi[%d]", then we can change x to "x[%d]".
        self.xi_sym = swiginac.matrix(nsd, 1, list(sfc_cell.xi[:nsd]))
        #self.x_sym  = symbols(("x[%d]" % i for i in range(nsd)))
        x_sym = symbols(("x%d" % i for i in range(nsd)))
        x_expr = [self.x0_sym[i] + sum(self.G_sym[i,j]*self.xi_sym[j] for j in range(nsd))
                  for i in range(nsd)] # FIXME: Transpose?
        self.x_sym  = swiginac.matrix(nsd, 1, x_sym)
        self.x_expr = swiginac.matrix(nsd, 1, x_expr)

        # Quadrature variables (local coordinates are self.xi_sym)
        self.quad_weight_sym = symbol("quad_weight")
        self.D_sym = symbol("D")
        self.D_expr = self.quad_weight_sym * self.detG_sym

        # Facet mapping
        #self.facet_G_sym = FIXME
        #self.facet_G_expr[facet] = FIXME

        # Facet normal expressions with sign scaling for inverted cells
        self.n_tmp_sym = symbolic_vector(nsd, "ntmp")
        self.n_tmp_expr = [[sum(self.G_sym[j,i]*n[j] for j in range(nsd))
                            for i in range(nsd)]
                           for n in sfc_cell.facet_n]

        self.nnorm_sym = symbol('nnorm')
        self.nnorm_expr = swiginac.sqrt(sum(self.n_tmp_sym[i]**2 for i in range(nsd)))

        self.n_sym = symbolic_vector(nsd, "n")
        self.n_expr = [(self.n_tmp_sym[i]/self.nnorm_sym).evalm() for i in range(nsd)]

        self.facet_D_sym = symbol("facet_D")
        self.facet_D_expr = build_facet_D(sfc_cell, self.G_sym)

def build_facet_D(sfc_cell, G_sym):
    sh = sfc_cell.shape
    fp = sfc_cell.facet_polygons
    if sh == "interval":
        return build_facet_D_interval(fp, G_sym)
    elif sh == "triangle":
        return build_facet_D_triangle(fp, G_sym)
    elif sh == "tetrahedron":
        return build_facet_D_tetrahedron(fp, G_sym)
    elif sh == "quadrilateral":
        return build_facet_D_quadrilateral(fp, G_sym)
    elif sh == "hexahedron":
        return build_facet_D_hexahedron(fp, G_sym)

def build_facet_D_interval(facet_polygons, G_sym):
    one = swiginac.numeric(1)
    return [one]*len(facet_polygons)

def build_facet_D_triangle(facet_polygons, G_sym):
    facet_D_expr = []
    for facet, facet_polygon in enumerate(facet_polygons):
        # facet_polygon is a line
        v0 = G_sym * list_to_vector(facet_polygon.vertex(0))
        v1 = G_sym * list_to_vector(facet_polygon.vertex(1))
        v  = sub(v1, v0)
        D = swiginac.sqrt( inner(v, v) ) # |v|
        if facet == 0: # diagonal facet
            D = D / swiginac.sqrt(2) # FIXME: Is this right?
        # TODO: Should in general use determinant of mapping
        #       instead of this linear approximation
        facet_D_expr.append(D)
    return facet_D_expr

def build_facet_D_tetrahedron(facet_polygons, G_sym):
    facet_D_expr = []
    for facet, facet_polygon in enumerate(facet_polygons):
        # facet_polygon is a triangle
        v0 = facet_polygon.vertex(0)
        v1 = facet_polygon.vertex(1)
        v2 = facet_polygon.vertex(2)
        # map local reference coordinates to get global cell coordinates
        v0 = G_sym * list_to_vector(v0)
        v1 = G_sym * list_to_vector(v1)
        v2 = G_sym * list_to_vector(v2)

        # vx[i][j] = component j of coordinate of vertex i in cell
        #vx_sym[i][j]

        # A = |a x b|/2, D = A/(1/2) = |a x b|,
        # with A = global triangle area, and
        # D being A over the reference area 1/2
        c = cross(sub(v1, v0), sub(v2, v0))
        D = swiginac.sqrt( inner(c, c) ) # |c|
        if facet == 0: # diagonal facet
            D = D / swiginac.sqrt(3) # FIXME: Is this right?
        # TODO: Should in general use determinant of mapping
        #       instead of this linear approximation
        facet_D_expr.append(D)
    return facet_D_expr

def build_facet_D_quadrilateral(facet_polygons, G_sym):
    facet_D_expr = []
    for facet, facet_polygon in enumerate(facet_polygons):
        # facet_polygon is a line
        v0 = G_sym * list_to_vector(facet_polygon.vertex(0))
        v1 = G_sym * list_to_vector(facet_polygon.vertex(1))
        v  = sub(v1, v0)
        D  = swiginac.sqrt(inner(v,v))
        # TODO: Should in general use determinant of mapping
        #       instead of this linear approximation
        facet_D_expr.append(D)
    return facet_D_expr

def build_facet_D_hexahedron(facet_polygons, G_sym):
    facet_D_expr = []
    for facet, facet_polygon in enumerate(facet_polygons):
        # facet_polygon is a quadrilateral
        v0 = G_sym * list_to_vector(facet_polygon.vertex(0))
        v1 = G_sym * list_to_vector(facet_polygon.vertex(1))
        v2 = G_sym * list_to_vector(facet_polygon.vertex(2))
        v3 = G_sym * list_to_vector(facet_polygon.vertex(3))
        # compute midpoints
        m0 = (v0 + v1)/2
        m1 = (v1 + v2)/2
        m2 = (v2 + v3)/2
        m3 = (v3 + v0)/2
        # area is length of cross product of vectors between opposing midpoints
        c = cross(sub(m2, m0), sub(m1, m3))
        D = swiginac.sqrt( inner(c, c) )
        # TODO: Should in general use determinant of mapping
        #       instead of this linear approximation
        facet_D_expr.append(D)
    return facet_D_expr
